import logging
import shelve

from homelab.control.utils import get_cache_path

USE_DNS = "use_dns"

instance = None

def getConfig():
    global instance
    if(instance == None):
        instance = Config()
    return instance


class Config:
    def __init__(self):
        with shelve.open(get_cache_path() + '/config.db', writeback=True) as db:
            if ("config" in db):
                self.config = db["config"]
            else:
                self.config = self.getDefaultConfig()
        logging.info("restored config: " + str(self.config))

    def getDefaultConfig(self):
        defconf = {}
        defconf[USE_DNS] = 1
        return defconf

    def getConfig(self):
        return self.config


"""
Copyright 2019, Nils Wenzler, All rights reserved.
nils.wenzler@crowgames.de
"""
import logging
import os
from pathlib import Path
from http.server import HTTPServer, BaseHTTPRequestHandler


class StaticServer(BaseHTTPRequestHandler):

    def do_GET(self):
        root = os.path.join(str(Path(os.path.dirname(os.path.realpath(__file__))).parent), 'static')

        print(root)
        # print(self.path)
        if self.path == '/':
            filename = root + '/index.html'
        else:
            filename = root + self.path

        self.send_response(200)
        if filename[-4:] == '.css':
            self.send_header('Content-type', 'text/css')
        elif filename[-5:] == '.json':
            self.send_header('Content-type', 'application/javascript')
        elif filename[-3:] == '.js':
            self.send_header('Content-type', 'application/javascript')
        elif filename[-4:] == '.ico':
            self.send_header('Content-type', 'image/x-icon')
        else:
            self.send_header('Content-type', 'text/html')
        self.end_headers()
        try:
            with open(filename, 'rb') as fh:
                html = fh.read()
                # html = bytes(html, 'utf8')
                self.wfile.write(html)
        except Exception as e:
            logging.exception(str(e))


def run(server_class=HTTPServer, handler_class=StaticServer, port=7464):
    server_address = ('', port)
    httpd = server_class(server_address, handler_class)
    logging.info('Starting httpd on port {}'.format(port))
    httpd.serve_forever()


"""
Copyright 2019, Nils Wenzler, All rights reserved.
nils.wenzler@crowgames.de
"""
import logging
import platform
import shelve
import socket
from uuid import getnode as get_mac

from scapy.layers.l2 import Ether, ARP
from scapy.sendrecv import srp

from homelab.analysis import nmap
from homelab.analysis.networkscanner import getNetworkScanner

from appdirs import *

from homelab.control.utils import get_cache_path

instance = None

def getDeviceLibrary():
    global instance
    if(instance == None):
        instance = DeviceLibrary()
    return instance

class DeviceLibrary:

    device_list = []
    history = []

    def __init__(self):
        with shelve.open(get_cache_path()+'/device.db', writeback=True) as db:
            if ("devices" in db):
                self.device_list = db["devices"]
            else:
                self.device_list = []
            if ("history" in db):
                self.history = db["history"]
            else:
                self.history = []
        logging.info("restored device list: "+str(self.device_list))

    def save_device_list(self):
        with shelve.open(get_cache_path()+'/device.db', writeback=True) as db:
            db["devices"] = self.device_list

    def save_scan_result(self, scan_result):
        self.history.append(scan_result)
        with shelve.open(get_cache_path()+'/device.db', writeback=True) as db:
            db["history"] = self.history

    def get_device_by_ip(self, ip):
        for device in self.device_list:
            if device["ip"] == ip:
                return device
        return None

    def arp_scan_network_for_devices(self):
        """ Returns a list of devices which are currently active on the network """
        default_gateway = getNetworkScanner().get_default_gateway()
        logging.info("[*] Scanning network for devices (this takes up to 30 seconds)")
        ip = default_gateway + "/24"

        answered_list = srp(Ether(dst="ff:ff:ff:ff:ff:ff") / ARP(pdst=ip), timeout=2, retry=10)[0]

        for element in answered_list:
            if element[1].psrc != default_gateway:
                device = self.get_device_by_ip(element[1].psrc)
                if(device == None):
                    device = {}
                    self.device_list.append(device)
                device["ip"] = element[1].psrc
                device["mac"] = element[1].hwsrc
                device["scanning"] = 0

        self.save_device_list()
        return self.device_list

    def nmap_scan_and_fingerprint(self):
        print("[*] Scanning network for devices (this takes up to 60 seconds)")
        default_gateway = getNetworkScanner().get_default_gateway()
        cidr = default_gateway + "/24"
        nm = nmap.PortScanner()
        # for name detection
        nm.scan(cidr, arguments='-sV -T4 -O -F --version-light')
        for ip in nm._scan_result['scan']:
            if('mac' in nm[ip]['addresses']):
                client_dict = {"ip": ip, "mac": nm[ip]['addresses']['mac'], "scanning": 0}
                already_found = False
                for device in self.device_list:
                    if(device["ip"] == ip):
                        client_dict = device
                        already_found = True
                if(len(nm[ip]['hostnames'])>0):
                    client_dict["name"] = nm[ip]['hostnames'][0]['name'].replace(".lan","")
                if('portsused' in nm[ip]):
                    client_dict["ports"] = nm[ip]['portsused']
                if ('tcp' in nm[ip]):
                    client_dict["tcp"] = nm[ip]['tcp']
                if ('osmatch' in nm[ip]):
                    client_dict["osmatch"] = nm[ip]['osmatch']

                if(not already_found):
                    self.device_list.append(client_dict)
            else:
                own_ips = [ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith("127.")][
                          : 1]
                #TODO make it work
                if (ip in own_ips or True):
                    client_dict = {"ip": ip, "mac": str(get_mac()), "scanning": 0}
                    if (len(nm[ip]['hostnames']) > 0):
                        client_dict["name"] = nm[ip]['hostnames'][0]['name'].replace(".lan", "")
                    if ('portsused' in nm[ip]):
                        client_dict["ports"] = nm[ip]['portsused']
                    if ('tcp' in nm[ip]):
                        client_dict["tcp"] = nm[ip]['tcp']
                    if ('osmatch' in nm[ip]):
                        client_dict["osmatch"] = nm[ip]['osmatch']
                else:
                    print("WARNING: a device ("+ip+") has been ignored because no MAC address was found")

        del nm
        self.save_device_list()
        return self.device_list

    def enhance_device_list(self, device_list):
        for device in device_list:
            os = self.fingerprint_device(device["ip"])
            if("hostnames" in os and len(os["hostnames"])>0):
                device["name"] = os["hostnames"][0]["name"].replace(".lan","")

    def fingerprint_device(self, device):
        """
        returns information concerning device and OS for a given IP
        :param ip: to scan (schould be reachable)
        :return: {"hostnames": [{"name":value}]}
        """

        # check whether scanning device
        own_ips = [ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith("127.")][: 1]
        if(device["ip"] in own_ips):
            return {"hostnames": [{"name":platform.node()}]}

        # otherwise perform scan
        try:
            logging.info("trying to perform deep scan and fingerprinting on "+device["ip"])
            nm = nmap.PortScanner()
            # for name detection
            #nm.scan(ip, arguments='-A')
            # for OS detection
            #nm.scan(device["ip"], arguments='-p 1-65535 -T4 -A -v ')
            nm.scan(device["ip"], arguments='-T4 -A -v')
            if not (nm[device["ip"]]["status"] == "down"):
                logging.info("deep scan and fingerprinting successfull for "+device["ip"])
                sc_res = nm[device["ip"]]
                if (len(sc_res['hostnames']) > 0):
                    device["name"] = sc_res['hostnames'][0]['name'].replace(".lan", "")
                if ('portsused' in sc_res):
                    device["ports"] = sc_res['portsused']
                if ('tcp' in sc_res):
                    device["tcp"] = sc_res['tcp']
                if ('osmatch' in sc_res):
                    device["osmatch"] = sc_res['osmatch']
                if ('vendor' in sc_res):
                    device["vendor"] = sc_res['vendor']
                device["fingerprint"] = 1
                self.save_device_list()
            else:
                logging.info("deep scan and fingerprinting failed (down) for " + device["ip"])
            return device
        except KeyError:
            print(KeyError)
            return {}


if __name__ == "__main__":
    test = DeviceLibrary()
    #res = test.fingerprint_device('192.168.8.113')
    res = test.fingerprint_device({"ip": "192.168.8.1"})
    print(res)


"""
Copyright 2019, Nils Wenzler, All rights reserved.
nils.wenzler@crowgames.de
"""
import copy
import http
import logging
import socket
import time

from scapy.layers.dns import DNSRR
from scapy.layers.inet import IP

from homelab.analysis.ipthreatdatabase import getIPThreatDatabase
from homelab.analysis.networkscanner import getNetworkScanner
from homelab.control.config import getConfig, USE_DNS

instance = None

def getTrafficAnalyzer():
    global instance
    if(instance == None):
        instance = TrafficAnalyzer()
    return instance

class TrafficAnalyzer:
    tot_p_count = 0
    ana_res = {}

    def analyze(self, packets_test, device_list):
        #own_ip = get('https://api.ipify.org').text

        self.ana_res = {}

        for device in device_list:
            if(device["ip"] not in self.ana_res):
                self.ana_res[device["ip"]] = {}

        default_gateway = getNetworkScanner().get_default_gateway()
        cidr = default_gateway + "/24"

        logging.info("[*] Start traffic analysis")
        logging.info(" |--> IP traffic analysis")
        logging.info("packets: "+str(len(packets_test)))

        # acclumerate IP packet information
        self.tot_p_count += len(packets_test)
        for pkt in packets_test:
            """if HTTP in pkt:
                http_layer = pkt.getlayer(HTTP)
                pkt.show()"""

            if IP in pkt:
                send = True
                ip_src = pkt[IP].src
                ip_dst = pkt[IP].dst
                foreign_ip = None
                local_ip = None

                if (ip_src in self.ana_res):
                    foreign_ip = ip_dst
                    local_ip = ip_src
                    send = True

                if (ip_dst in self.ana_res):
                    foreign_ip = ip_src
                    local_ip = ip_dst
                    send = False

                if(local_ip != None and foreign_ip != None):
                    if (foreign_ip not in self.ana_res[local_ip]):
                        if(send):
                            self.ana_res[local_ip][foreign_ip] = {"snd": 1, "rcv":0, "bsnd": len(pkt), "brcv":0, "domains": [], "threats": []}
                        else:
                            self.ana_res[local_ip][foreign_ip] = {"snd": 0, "rcv":1, "bsnd": len(pkt), "brcv":0, "domains": [], "threats": []}
                    else:
                        if(send):
                            self.ana_res[local_ip][foreign_ip]["snd"] += 1
                            self.ana_res[local_ip][foreign_ip]["bsnd"] += len(pkt)
                        else:
                            self.ana_res[local_ip][foreign_ip]["rcv"] += 1
                            self.ana_res[local_ip][foreign_ip]["brcv"] += len(pkt)


        # perform IP check
        #collect all ips
        all_ips = set()

        for key in self.ana_res.keys():
            for ip in self.ana_res[key]:
                all_ips.add(ip)

        ip_threats = getIPThreatDatabase().check_IPs(all_ips)

        for key in self.ana_res.keys():
            for ip in self.ana_res[key]:
                # Check for malicious ips
                if(ip in ip_threats):
                    self.ana_res[key][ip]["threats"].append(ip_threats[ip])

                # Perform reverse dns lookup
                if(len(self.ana_res[key][ip]["domains"])==0):
                    try:
                        reversed_dns = socket.gethostbyaddr(ip)
                        if(len(reversed_dns)>0):
                            self.ana_res[key][ip]["domains"].append(reversed_dns[0])
                    except Exception:
                        logging.exception("no reverse dns for "+ip)

        logging.info(" |--> DNS traffic analysis")
        if(getConfig().getConfig()[USE_DNS]>0):
            logging.info("      DNS traffic analysis enabled")
            for pkt in packets_test:
                if pkt.haslayer(DNSRR):
                    # If the answer is a DNSRR, print the name it replied with.
                    if isinstance(pkt.an, DNSRR):
                        for key in self.ana_res.keys():
                            if(pkt.an.rdata in self.ana_res[key]):
                                self.ana_res[key][pkt.an.rdata]["domains"].append(pkt.an.rrname.decode("utf-8"))
                                # TODO: perform malicious dns lookup
        else:
            logging.info("      DNS traffic analysis disabled")

        logging.info("[*] Traffic analysis completed")

        for device in device_list:
            device["connections"] = self.ana_res[device["ip"]]
            getNetworkScanner().add_locations_to_device(device)
        return {"devices":copy.deepcopy(device_list), "cidr": cidr, "time":time.time()}



#if __name__ == "__main__":
#    test = getTrafficAnalyzer()
#    packets = rdpcap('full_capture.pcap')
#    print(test.analyze(packets, [{"ip":"192.168.8.113"},{"ip":"192.168.8.101"},{"ip":"192.168.8.1"}]))


"""
Copyright 2019, Nils Wenzler, All rights reserved.
nils.wenzler@crowgames.de
"""

import logging
import os

instance = None

def getIPThreatDatabase():
    global instance
    if(instance == None):
        instance = IPTreatDatabase()
    return instance

class IPTreatDatabase:
    ips = {}
    def __init__(self):
        logging.info("[*] Initializing threat database ")
        # recursively traverse threats folder
        """for cur, _dirs, files in os.walk("threats"):
            pref = 'threats/'
            head, tail = os.path.split(cur)
            while head:
                pref += tail+'/'
                head, _tail = os.path.split(head)
            for file in files:
                if file.endswith(".ipset"):
                    filepath = pref+file
                    print(".", end='')
                    with open(filepath, "r") as ins:
                        for line in ins:
                            if "#" not in line:
                                line = line.strip('\n\r ')
                                if(line not in self.ips):
                                    self.ips[line] = {"lists":[]}
                                self.ips[line]["lists"].append(filepath)
        print(" done.")
        print("Loaded "+str(len(self.ips.keys()))+" unique IPs")"""

    def check_IPs(self, ips):
        result = {}
        """for cur, _dirs, files in os.walk("threats"):
            pref = 'threats/'
            head, tail = os.path.split(cur)
            while head:
                pref += tail+'/'
                head, _tail = os.path.split(head)
            for file in files:
                if file.endswith(".ipset"):
                    filepath = pref+file
                    print(".", end='')
                    with open(filepath, "r") as ins:
                        for line in ins:
                            if "#" not in line:
                                line = line.strip('\n\r ')
                                if(line in ips):
                                    if(line not in result):
                                        result[line] = {"lists":[]}
                                    result[line]["lists"].append(filepath)
                                    #self.ips[line] = {"lists":[]}
                                    #self.ips[line]["lists"].append(filepath)"""
        return result

if __name__ == "__main__":
    test = getIPThreatDatabase()
    print(test.check_IP("91.215.103.64"))


import asyncio
import json
import logging
import os
import threading
import time

import websockets
from pympler import muppy, summary

from homelab.analysis.devicelibrary import getDeviceLibrary
from homelab.analysis.networkscanner import getNetworkScanner
from homelab.control.jobscheduler import getJobScheduler


class WebSocketServer:



    def print_memory(self):
        all_objects = muppy.get_objects()
        sum1 = summary.summarize(all_objects)
        # Prints out a summary of the large objects
        summary.print_(sum1)
        # Get references to certain types of objects such as dataframe
        #dataframes = [ao for ao in all_objects if isinstance(ao, pd.DataFrame)]
        #for d in dataframes:
        #    print(d.columns.values)
        #    print(len(d))

    def scan_trigger(self, websocket):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(self.scan(websocket))

    async def scan(self, websocket):
        cnt = 0
        default_gateway = getNetworkScanner().get_default_gateway()
        cidr = default_gateway + "/24"
        last_update = 0
        while(True):

            if(last_update < getJobScheduler().getLastUpdate()):
                last_update = getJobScheduler().getLastUpdate()
                scan_result = {"devices":getDeviceLibrary().device_list, "cidr": cidr}

                await websocket.send("{\"type\":\"scan_result\", \"payload\":" + json.dumps(scan_result) + "}")
                await websocket.send("{\"type\":\"history\", \"payload\":" + json.dumps(getDeviceLibrary().history) + "}")

            time.sleep(1)

    async def handle_connection(self, websocket, path):
        try:
            scanning_thread = threading.Thread(target=self.scan_trigger,
                                               args=(websocket,), name="Regular Transmit")
            scanning_thread.start()
            while(True):
                message = await websocket.recv()
                data = json.loads(message)
                if (data["type"] == "kill"):
                    os._exit(1)
        except Exception:
            logging.info("websocket died")



    def __init__(self):
        logging.info("[*] Start local websocket server for app interaction")
        self.start_server = websockets.serve(self.handle_connection, '192.168.8.107', 7463)

        asyncio.get_event_loop().run_until_complete(self.start_server)
        asyncio.get_event_loop().run_forever()





import logging
import os

from appdirs import user_data_dir


def get_cache_path():
    path = user_data_dir("HomeLab", "NilsWe")
    logging.info("using " + path + " as local cache directory")
    if not os.path.exists(path):
        os.makedirs(path)
    return path


import logging
import threading
import time
from enum import Enum

from homelab.analysis.devicelibrary import getDeviceLibrary
from homelab.analysis.networkscanner import getNetworkScanner
from homelab.analysis.trafficanalyzer import getTrafficAnalyzer


class Job(Enum):
    FINGERPRINT = 1
    DEVICE_SEARCH = 2
    NETWORK_CAPTURE = 3
    NETWORK_ANALYZE = 4
    IDLE = 5

instance = None

def getJobScheduler():
    global instance
    if(instance == None):
        instance = JobScheduler()
    return instance

class JobScheduler:
    curJob = Job.IDLE
    lastScan = 1
    lastSearch = 1
    thread = None

    def getLastUpdate(self):
        return max(self.lastScan, self.lastSearch)

    def __init__(self):
        self.thread = threading.Thread(target=self.work, name="JobScheduler")
        self.thread.start()

    def work(self):
        while(True):
            if(self.curJob == Job.IDLE):
                # if no device found go for device search
                if(len(getDeviceLibrary().device_list)==0):
                    self.startDeviceSearch()
                    continue

                # if lastScan more then 1 minutes ago scan
                if(time.time() - self.lastScan > 60*1):
                    self.startNetworkScan()
                    continue

                # if lastScan more then 5 minutes ago scan
                if (time.time() - self.lastSearch > 60 * 5):
                    self.startDeviceSearch()
                    continue

                # if devices exist that have not been fingerprinted fingerprint
                tmp_cnt = False
                for device in getDeviceLibrary().device_list:
                    if("fingerprint" not in device):
                        self.startDeviceFingerprint(device)
                        tmp_cnt = True
                        break
                if(tmp_cnt):
                    continue

                # otherwise idle for a minute
                self.startIdle()

    def startDeviceSearch(self):
        logging.info("[i] start device search job")
        self.curJob = Job.DEVICE_SEARCH
        device_list = getDeviceLibrary().arp_scan_network_for_devices()
        device_list = getDeviceLibrary().nmap_scan_and_fingerprint()
        self.lastSearch = time.time()
        self.curJob = Job.IDLE

    def startNetworkScan(self):
        logging.info("[i] start network scanning job")
        self.curJob = Job.NETWORK_CAPTURE
        getNetworkScanner().set_devices_scanning(getDeviceLibrary().device_list, 1)
        getNetworkScanner().print_device_list(getDeviceLibrary().device_list)
        packets = getNetworkScanner().sniff_devices(getDeviceLibrary().device_list, timeout=30)
        self.curJob = Job.NETWORK_ANALYZE
        getNetworkScanner().set_devices_scanning(getDeviceLibrary().device_list, 0)
        scan_result = getTrafficAnalyzer().analyze(packets, getDeviceLibrary().device_list)

        getDeviceLibrary().save_scan_result(scan_result)

        # remove packets from memory (can be huge!)
        del packets


        self.lastScan = time.time()
        self.curJob = Job.IDLE

    def startDeviceFingerprint(self, device):
        logging.info("[i] start device fingerpinting job")
        self.curJob = Job.FINGERPRINT
        getDeviceLibrary().fingerprint_device(device)
        logging.info("[i] finished device fingerpinting job")
        self.curJob = Job.IDLE

    def startIdle(self):
        logging.info("[i] start idling job")
        self.curJob = Job.IDLE
        time.sleep(60)
        self.curJob = Job.IDLE